/*  GNP.PRG 

This is a GAUSS program, written by Bruce E. Hansen.
It replicates Table 5 and Figures 3 and 4 from my paper
"The Grid Bootstrap and the Autoregressive Model".

If you have questions, please contact me at

Bruce E. Hansen
Department of Economics
Social Science Building
University of Wisconsin
Madison, WI 53706-1393
bhansen@ssc.wisc.edu

*/

load np[13,4]=np_index.dat;
load dat[]=nplosser.dat;

orig = 2;    /* set to 1 for original data, set to 2 for extended data */
t=2;
grid=200;
bq=9999;
c=.9;

output file = gnp.out reset;

i=7;

  d=np[i,.];
  if orig==1;
    y=dat[d[2]:d[3]-18];
    if i==4;  y=y[21:82];  endif;
  else;
    y=dat[d[2]:d[3]];
  endif;
  name = "GNP per Capita: 1869-1988";
  ar=d[4];
  call grid_boot(y,name,t,ar,grid,bq,c);

output off;

/*****************************************************************/

proc (0)=grid_boot(dat,name,t,ar,grid,bq,c);
local cc,n,y0,y,x,betahat,sehat,e,g_lu,b_lu,k,bk,sk,beta0,kk,xkk,xk,tstat,
i,bi,yk,bbi,phi,j,ee,datb,yb,xb,bb,sb,eb,qh,tk,g_l,g_u,g_lh,g_uh,
b_lh,b_uh,b_l,b_u,z,fmtar,tit,par,xname,kname,ff,pr,ff1,ff2,hname,rr,y00,
ai,ti,ei,d0,dt,yt;

  cc=((1-c)/2)|((c+1)/2);
  n=rows(dat);
  if t==0; yt=dat; endif;
  if t==1; yt=dat-meanc(dat);endif;
  if t==2; dt=ones(n,1)~seqa(1,1,n); yt=dat-dt*(dat/dt);endif;
  y0=yt[1:ar];
 {y,x,betahat,sehat,e}= est(dat,t,ar);
  g_lu=zeros(ar,2);
  b_lu=zeros(ar,2);
  k=1; do while k<=ar;
    bk=betahat[t+k];
    sk=sehat[t+k];
    beta0=seqa(bk-5*sk,10*sk/(grid-1),grid);
    kk=setdif(seqa(1,1,ar+t),k+t,1);
    xk=x[.,kk];
    tstat=zeros(bq,grid);
    i=1;do while i<=grid;
      bi=beta0[i];
      yk=y-x[.,k+t]*bi;
      ai=yk/xk;
      if ar>1;
        bbi=zeros(ar,1);
        bbi[k]=bi;
        bbi[setdif(seqa(1,1,ar),k,1)]=ai[1+t:ar+t-1];
        phi = (bbi[2:ar]|0)-(0|bbi[2:ar]);
        phi[1]=phi[1]+bbi[1];
      else;  phi=bi;
      endif;
      rr=polyroot((-rev(phi))|1);
      if minc(abs(rr))<=1;
        y00=zeros(ar,1);
      else;
        y00=y0;
      endif;
      j=1;do while j<=bq;
        datb=recserar(e[trunc(1+(n-ar)*rndu(n,1))],y00,phi);
       {yb,xb,bb,sb,eb}= est(datb,t,ar);
       tstat[j,i]=(bb[t+k]-bi)/sb[t+k];
      j=j+1;endo;
    i=i+1;endo;
    qh=kq_reg(tstat,beta0,cc);      
    tk=(bk-beta0)/sk;
    {g_l,g_u}=g_ends(beta0,tk,qh);
    g_lu[k,.]=g_l~g_u;
    g_lh=height(tk,beta0,g_l);
    g_uh=height(tk,beta0,g_u);
    b_lh=height(qh[.,2],beta0,bk);
    b_uh=height(qh[.,1],beta0,bk);
    b_l=bk-b_lh*sk;
    b_u=bk-b_uh*sk;
    b_lu[k,.]=b_l~b_u;
    z=zeros(grid,1);
    graphset;
    _pcolor=15;
    _psym=
    (bk~0~8~3~15~1~0)
    |(bk~b_uh~5~3~15~1~0)
    |(bk~b_lh~5~3~15~1~0)
    |(b_l~b_lh~2~3~15~1~0)
    |(b_u~b_uh~2~3~15~1~0)
    |(g_l~g_lh~1~3~15~1~0)
    |(g_u~g_uh~1~3~15~1~0);
    _pline=
    (1~4~b_l~b_lh~bk~b_lh~1~15~5)
    |(1~4~bk~b_lh~bk~b_uh~0~15~5)
    |(1~4~bk~b_uh~b_u~b_uh~0~15~5);
    _parrow=
    (g_l~g_lh~g_l~0~4~.2~01~15~1~6~2)
    |(g_u~g_uh~g_u~0~4~.2~01~15~1~6~2)
    |(b_u~b_uh~b_u~0~4~.2~11~15~1~4~5)
    |(b_l~b_lh~b_l~0~4~.2~11~15~1~4~5);
    _pdate="";
    fonts("complex simgrma");
    _plwidth=2;
    let _pltype = 6 6 3 1;
    fmtar = "AR(%*.*lf)";
    par = "\202r\201]" $+ ftos(k,"%*.*lf",1,0) $+ "[";
    tit = "" $+ "Figure " $+ ftos(2+k,"%*.*lf",1,0) $+ "\L" $+ name;
    title(tit);
    tit = "" $+ ftos(c*100,"%*.*lf",2,0) $+ "% Confidence Interval for " $+ par;
    xlabel(tit);
    ylabel("Bootstrap Quantiles");
    if k==1; 
      xtics(.7,1.05,.05,5);
      ytics(-4,4,2,4);
    elseif k==2;
      xtics(.1,.7,.1,10);
      ytics(-4,4,2,4);
    endif;
    xy(beta0,tk~z~qh[.,2 1]);
    _protate=1;
    if k==1;
      graphprt("-c=1 -cf=fig_3.eps");
    elseif k==2;
      graphprt("-c=1 -cf=fig_4.eps");
    endif;
    xy(beta0,tk~z~qh[.,2 1]);
    

  k=k+1; endo;
  xname="Y(t-1)";
  k=2; do while k<=ar;
    kname="" $+ "DY(t-" $+ ftocv(k,1,0) $+ ")";
    xname=xname|kname;
  k=k+1;endo;
  ff1 = "*.*lf"~10~10;
  ff2 = "*.*lf"~10~4;
  ff = ff1|(ones(6,1)*ff2); 
  $name;
  "Number of Gridpoints                    " grid;
  "Number of Bootstrap Replications " bq;
  "Confidence Level                          " c;
  "";"";
  hname="Variable"~"Estimate"~"S.E."~"Grid_L"~"Grid_U"~"Per-t_L"~"Per-t_U";
  pr = printfm(hname,zeros(1,7),ff1);"";
  "      ______________________________________________________________";
  pr = printfm(xname~betahat[t+1:t+ar]~sehat[t+1:t+ar]~g_lu~b_lu,0~ones(1,6),ff);
  "      ______________________________________________________________";
  "";"";
  "**********************************************************";
  "";"";

endp;


/*****************************************************************
PROC EST
Format: 	
	{y,x,betahat,sehat,e}=est(dat,t,ar);
Inputs:  
	dat	nx1 data vector
	t	trend order.  0 for no trend, 1 for constant, 2 for constant and trend
	ar	autoregressive order
Outputs:
	y	nx1 vector of dependent variable
	x	nx(ar+t) matrix of regressors
	betahat	OLS estimates
	sehat	OLS standard errors
	e	OLS residuals
*/
proc (5) = est(dat,t,ar);
local n,x,k,y,xxi,betahat,e,sehat;
  n=rows(dat);
  if t==0;
    x=dat[ar:n-1];
  elseif t==1;
    x=ones(n-ar,1)~dat[ar:n-1];
  elseif t==2;
    x=ones(n-ar,1)~seqa(1,1,n-ar)~dat[ar:n-1];
  endif;
  k=2; do while k<=ar;
    x=x~(dat[ar-k+2:n-k+1]-dat[ar-k+1:n-k]);
  k=k+1;endo;
  y=dat[1+ar:n];
  xxi = invpd(moment(x,0));
  betahat=xxi*(x'y);
  e=y-x*betahat;
  sehat=sqrt(diag(xxi)*(e'e)/(n-ar-t));
retp(y,x,betahat,sehat,e);
endp;
/**************************************************************/


/*****************************************************************
PROC QUANTILE
Format: 	
	q=quantile(x,c);
Inputs:  
	x	nxr data matrix
	c	px1 list of quantiles
Output:
	q	rxp matrix of quantiles
*/
proc quantile(x,c);
local b,r,xx,j,bc,i,q;
  b=rows(x); r=cols(x);xx=x;
  j=1; do while j<=r;
    xx[.,j]=sortc(x[.,j],1);
  j=j+1;endo;
  bc=b*c;
  i=floor(bc);
  i=i+(i .== 0);
  q=xx[i,.].*(i .== bc) + ((xx[i,.]+xx[i+1,.])/2).*(i .ne bc);
retp(q');
endp;
/**************************************************************/


/*****************************************************************
PROC EPI
Format: 	
	w=epi(a,h,leave);
Inputs:  
	a	nx1 vector of points to evaluate kernel
	h	scalar bandwidth
	leave	set to 1 to impose leave-one-out kernel estimation
Output:
	w	nxn matrix of kernel weights
*/
proc epi(a,h,leave);
local u,w;
  u=(a-a')/h;
  w=(1-u.^2).*(abs(u) .<= 1);
  if leave==1;
    w=diagrv(w,zeros(rows(a),1));
  endif;
  w=w./sumc(w');
retp(w);
endp;
/**************************************************************/


/*****************************************************************
PROC KQ_REG
Format: 	
	qh=kq_reg(y,x,c);
Inputs:  
	y	nxr data matrix
	x	rx1 vector of points to evaluate kernel regression
	c	px1 list of quantiles
Output:
	qh	rxp matrix of kernel-smoothed quantiles
*/
proc kq_reg(y,x,c);
local hn,hh,cn,check,q,hi,qh,h,ci;
  hn=5;
  cn=rows(c);
  hh=seqa(1.5,1,hn)*(x[2]-x[1]);
  check=zeros(hn,cn);
  q=quantile(y,c);
  hi=1; do while hi<=hn;
    h=hh[hi];
    if x[2]-x[1]<h;
      qh=epi(x,h,1)*q;
      check[hi,.]=sumc((q-qh).^2)';
     else;check[hi,.]=sumc(q.^2)';
     endif;
  hi=hi+1;endo;
  hi=minindc(check);
  h=hh[hi];
  qh=zeros(rows(x),cn);
  ci=1; do while ci<=cn;
    qh[.,ci]=epi(x,h[ci],0)*q[.,ci];
  ci=ci+1;endo;
retp(qh);
endp;
/**************************************************************/


/*****************************************************************
PROC CROSS
Format: 	
	{g1,g2}=cross(a,x,y);
Inputs:  
	a	rx1 list of gridpoints
	x	rx1 list, first line
	c	rx1 list, second line
Outputs:
	g1	list of upcrossings (second line intersects first from below)
	g2	list of downcrossings (second line intersects first from above)
*/
proc (2) = cross(a,x,y);
local ms,rn,r,d,s,li,al,bl,au,bu,g1,g2;
  ms=0;  ms=miss(ms,ms);
  rn=rows(a);
  r=(x.<y);
  d=0|(r[2:rn]-r[1:rn-1]);
  s=seqa(1,1,rn);
  li=selif(s,(d .== 1));
  if maxc(li) .ne ms;
    al=a[li-1];au=a[li];
    bl=x[li-1]-y[li-1];   bu=y[li]-x[li];
    g1=(bu.*al+bl.*au)./(bl+bu);
  else;
    g1=ms;
  endif;
  li=selif(s,(d .== -1));
  if maxc(li) .ne ms;
    al=a[li-1];au=a[li];
    bl=y[li-1]-x[li-1];   bu=x[li]-y[li];
    g2=(bu.*al+bl.*au)./(bl+bu);
  else;
    g2=ms;
  endif;
retp(g1,g2);
endp;
/**************************************************************/


/*****************************************************************
PROC G_ENDS
Format: 	
	{l,u}=g_ends(a0,ta,grid);
Inputs:  
	a0	rx1 list of gridpoints
	ta	rx1 test statistic line
	grid	rx2 lower and upper quantiles
Output:
	l	lower endpoint of confidence interval
	u	upper endpoint of confidence interval
*/
proc (2) = g_ends(a0,ta,grid);
local l1,l2,u1,u2,s,e1,e2,u,l;
  {l1,l2}=cross(a0,ta,grid[.,1]);
  {u1,u2}=cross(a0,ta,grid[.,2]);
  s=packr(sortc(l1|l2|u1|u2,1));
  l=minc(s);
  u=maxc(s);
  if l==minc(u2); l=minc(a0); endif;
  if l==minc(l1); l=minc(a0); endif;
  if u==maxc(l2); u=maxc(a0); endif;
  if u==maxc(u1); u=maxc(a0); endif;
retp(l,u);
endp;
/**************************************************************/

/*****************************************************************
PROC HEIGHT
Format: 	
	qh=height(q,a,l);
Inputs:  
	q	rx1 heights at gridpoints
	a	rx1 gridpoints
	l	scalar, point to evaluate q
Output:
	qh	scalar, value of q at point l
*/
proc height(q,a,l);
local l1,l2,q1,q2,w1,w2,qh;
  l2=maxindc(a.>l);
  l1=l2-1;
  q1=q[l1];
  q2=q[l2];
  w1=a[l2]-l;
  w2=l-a[l1];
  qh=(w1*q1+w2*q2)/(w1+w2);
retp(qh);
endp;
/**************************************************************/